from abc import ABCMeta, abstractmethod
import json

from flippy.core import FeatureEncoder, FeatureName, Feature, Gate


class BaseBackend(metaclass=ABCMeta):
    """
    Flipper established a protocol for adapters to implement.
    This is the Pythonic implementation of it.
    
    https://www.flippercloud.io/docs/adapters/new
    """
    @abstractmethod
    def features(self) -> set[FeatureName]:
        "Get the set of known features."
        pass

    @abstractmethod
    def add(self, feature: FeatureName) -> bool:
        "Add a feature to the set of known features."
        pass

    @abstractmethod
    def remove(self, feature: FeatureName) -> bool:
        "Remove a feature from the set of known features."
        pass

    @abstractmethod
    def clear(self, feature: FeatureName) -> bool:
        "Clear all gate values for a feature."
        pass

    @abstractmethod
    def get(self, feature: FeatureName) -> Feature:
        "Get all gate values for a feature."
        pass

    @abstractmethod
    def enable(self, feature: FeatureName, gate: Gate, thing: str | int | None=None) -> bool:
        "Enable a gate for a thing."
        pass

    @abstractmethod
    def disable(self, feature: FeatureName, gate: Gate, thing: str | int | None=None) -> bool:
        "Disable a gate for a thing."
        pass

    @abstractmethod
    def get_multi(self, features: list[FeatureName]) -> list[Feature]:
        "Get all gate values for several features at once."
        # default implementation; feel free to use or override
        values = []
        for feature in features:
            values.append(self.get(feature))
        return values

    @abstractmethod
    def get_all(self) -> list[Feature]:
        "Get all gate values for all features at once."
        # default implementation; feel free to use or override
        values = []
        for feature in self.features():
            values.append(self.get(feature))
        return values
    
    @abstractmethod
    def to_json(self) -> str:
        "Produce a JSON-formatted string containing state for all features."
        # default implementation; feel free to use or override
        features = {f.key: f for f in self.get_all()}
        return json.dumps(features, cls=FeatureEncoder, separators=(',', ':'))

    @abstractmethod
    def from_json(self, new_state: str) -> None:
        "Clear current state and replace with state from a JSON-formatted string."
        pass

    # dict implementation
    # note, there's no setting of values because that doesn't fit the
    # dict contract cleanly
    def __len__(self):
        return len(self.features())
    
    def __getitem__(self, key: FeatureName):
        return self.get(key)
    
    def __delitem__(self, key: FeatureName):
        return self.remove(key)
    
    def __iter__(self):
        return iter(self.features())
    
    def __contains__(self, key: FeatureName):
        return key in self.features()
    # end dict implementation

BaseBackend.register(dict)



from dataclasses import dataclass
from flippy.backends.base import BaseBackend
from flippy.core import FeatureName, Gate
from flippy.exceptions import FeatureNotFound

ACTOR_IF_NO_TARGET = "anonymous"


@dataclass
class FeatureState:
    key: str
    boolean: bool | None
    actors: list[str]
    groups: list[str]
    percent_actors: int | None
    percent_time: int | None


class Flippy:
    """
    The `Flippy` object is the key interface to everything in Flippy.

    You can get ahold of one by passing in a backend, like this:
    
    ```python
    from flippy.backends import MemoryBackend
    from flippy import Flippy

    f = Flippy(MemoryBackend())
    ```
    """
    def __init__(self, backend: BaseBackend):
        """
        Available backends include:
        - `flippy.backends.MemoryBackend`
        - `flippy.backends.DjangoBackend`
        - `flippy.backends.FlipperCloudBackend`
        """
        self._backend = backend
    
    def is_enabled(self, feature: FeatureName, target = None) -> bool:
        """
        Checks whether a particular feature is enabled for a particular object
        (or globally, if you don't pass in a target).

        ```python
        # globally enabled?
        flippy.is_enabled('my_cool_feature')

        # enabled for this Django user?
        # (actually this will work for any object with a `pk` or `id` property)
        user = User.objects.get(pk=1)
        flippy.is_enabled('my_cool_feature', user)
        ```

        Targets are munged into "classname;id" so that they can round-trip through
        backends like Flipper Cloud just as well as they round-trip the Django ORM.
        This means that if you change the class name, objects will no longer be
        recognized if they were previously entered into Flippy. You can implement
        a `get_flipper_id` method on your object to precisely control this munging.
        We strongly suggest you use a similar scheme which "namespaces" IDs like
        the default implementation.

        ```python
        # enabled for an object with a get_flipper_id() method?
        @dataclass
        class Fruit:
            name: str
            produce_lookup_code: str

            # without this method, the default would be "Fruit;<id>", which you
            # may not want if you intend to also use produce lookup codes for,
            # say, vegetables. Instead we'll inject our own "PLU;" namespace.
            def get_flipper_id(self):
                return f"PLU;{self.produce_lookup_code}"
        
        banana = Fruit(name='banana', produce_lookup_code='4011')
        flippy.is_enabled('my_cool_feature', banana)
        ```
        """
        try:
            f = self._backend.get(feature)
        except FeatureNotFound:
            return False

        # if the boolean gate is on or off, that's final
        match f.state:
            case 'on':
                return True
            case 'off':
                return False

        # if the feature is conditional and no target was given, use a constant
        if target is None:
            actor = ACTOR_IF_NO_TARGET
        else:
            actor = self._to_flipper_id(target)

        if any([
            f.actors_gate.is_open(actor, feature),
            f.groups_gate.is_open(actor, feature),
            f.percentage_of_actors_gate.is_open(actor, feature),
            f.percentage_of_time_gate.is_open(actor, feature),
            f.expression_gate.is_open(actor, feature),
        ]):
            return True
        
        # TODO: special check whether actor is a member of an enabled group
        # (using the Django authentication system)
        ...
        
        return False
    
    def create(self, feature: FeatureName) -> bool:
        """
        Create a new feature flag.

        You should always do this, either using the Flippy frontend or
        through some backend-specific mechanism. If you try to set a feature
        flag's state without explicitly creating it, it's up to the backend
        whether to support that or raise an exception.
        """
        return self._backend.add(feature)
    
    def get_all_feature_names(self) -> set[FeatureName]:
        """
        Get a list of all known feature flag names.
        """
        return {f.key for f in self._backend.get_all()}
    
    def feature_exists(self, feature: FeatureName) -> bool:
        """
        Check if a specific feature name is known to this backend.
        """
        try:
            self._backend.get(feature)
            return True
        except FeatureNotFound:
            return False
    
    def get_feature_state(self, feature: FeatureName) -> FeatureState:
        """
        Get a lot of detail about exactly how a feature flag is enabled.

        This is not recomended as a general practice for checking feature
        flag state. Instead, use `Flippy.is_enabled` for those uses.
        This method is provided so that it's possible to build a frontend
        (web, CLI, etc.) for controlling feature flags.
        """
        f = self._backend.get(feature)
        return FeatureState(
            key=f.key,
            boolean=f.boolean_gate.value,
            actors=f.actors_gate.value,
            groups=f.groups_gate.value,
            percent_actors=f.percentage_of_actors_gate.value,
            percent_time=f.percentage_of_time_gate.value,
        )
    
    def enable(self, feature: FeatureName) -> None:
        """
        Globally enable a particular feature flag. This means it's on for
        everyone, all the time, no matter what other states are set.
        """
        try:
            f = self._backend.get(feature)
            self._backend.enable(f.key, Gate.Boolean)
        except FeatureNotFound:
            pass

    def enable_actor(self, feature: FeatureName, target) -> None:
        """
        Enable a feature for a particular actor. Usually this would be a user,
        but there may be special situations where you want to use other kinds
        of actors.
        """
        try:
            f = self._backend.get(feature)
            self._backend.enable(f.key, Gate.Actors, self._to_flipper_id(target))
        except FeatureNotFound:
            pass
    
    def enable_group(self, feature: FeatureName, target) -> None:
        """
        Enable a feature for an entire group of actors.
        
        At the present moment, this is not plumbed into the Django notion of
        groups, so this method is little different _in practice_ from
        `Flippy.enable_actor`. (You have to explicitly get a reference to the
        actual group object and pass it as the `target` to `Flippy.is_enabled`.)
        The design goal and future intention, though, is that calling
        `Flippy.is_enabled` on a user who belongs to that group will return
        True if that user is a member of an enabled group.
        """
        try:
            f = self._backend.get(feature)
            self._backend.enable(f.key, Gate.Groups, self._to_flipper_id(target))
        except FeatureNotFound:
            pass
    
    def enable_percentage_of_actors(self, feature: FeatureName, percentage: int) -> None:
        """
        Enable a feature for a percentage of all actors.

        Backends are expected to use a stable mechanism to opt a particular
        actor into a particular feature flag at any percentage threshold. That is,
        if an actor is enabled at % = 10, then that same user must still be enabled
        at % = 11, 12, 13, ..., 100. Likewise, if an actor is _not_ enabled at
        % = 20, then that same actor must not be enabled at % = 19, 18, ..., 0.
        """
        assert isinstance(percentage, int)
        assert 0 <= percentage <= 100

        try:
            f = self._backend.get(feature)
            self._backend.enable(f.key, Gate.PercentageOfActors, percentage)
        except FeatureNotFound:
            pass

    def enable_percentage_of_time(self, feature: FeatureName, percentage: int) -> None:
        """
        Enable a feature for a percentage of all lookups.

        On average, a feature marked for "50% of time" should come up as
        "enabled" about as often as it comes up "disabled". This isn't great
        for features which modify user experience (since subsequent page loads
        might give the opposite result) but could be useful for safely rolling
        out things like performance optimizations.
        """
        assert isinstance(percentage, int)
        assert 0 <= percentage <= 100

        try:
            f = self._backend.get(feature)
            self._backend.enable(f.key, Gate.PercentageOfTime, percentage)
        except FeatureNotFound:
            pass

    def disable(self, feature: FeatureName) -> None:
        """
        Turn off global enablement of a particular feature flag. This means it's
        on for only those targets specifically opted into it (actors, groups, % of
        actors, % of time).

        If you intend to totally disable the feature for _everyone_, instead of
        this method, you want `Flippy.clear`.
        """
        try:
            f = self._backend.get(feature)
            self._backend.disable(f.key, Gate.Boolean)
        except FeatureNotFound:
            pass

    def disable_actor(self, feature: FeatureName, target) -> None:
        """
        Remove an actor from the list of those enabled for the feature.
        """
        try:
            f = self._backend.get(feature)
            self._backend.disable(f.key, Gate.Actors, self._to_flipper_id(target))
        except FeatureNotFound:
            pass
    
    def disable_group(self, feature: FeatureName, target) -> None:
        """
        Remove a group from the list of those enabled for the feature.
        """
        try:
            f = self._backend.get(feature)
            self._backend.disable(f.key, Gate.Groups, self._to_flipper_id(target))
        except FeatureNotFound:
            pass
    
    def disable_percentage_of_actors(self, feature: FeatureName) -> None:
        """
        Disable the percentage-of-actors gate for a feature. This is the same
        as setting it to 0.
        """
        try:
            f = self._backend.get(feature)
            self._backend.disable(f.key, Gate.PercentageOfActors)
        except FeatureNotFound:
            pass

    def disable_percentage_of_time(self, feature: FeatureName) -> None:
        """
        Disable the percentage-of-time gate for a feature. This is the same
        as setting it to 0.
        """
        try:
            f = self._backend.get(feature)
            self._backend.disable(f.key, Gate.PercentageOfTime)
        except FeatureNotFound:
            pass
    
    def clear(self, feature: FeatureName) -> bool:
        """
        Globally disable a particular feature flag. This is a destructive action
        which clears all state associated with the flag, setting it back to
        disabled for _everyone_ including those you had previously opted in.
        """
        return self._backend.clear(feature)

    def destroy(self, feature: FeatureName) -> bool:
        """
        Tell the backend to completely delete this feature flag. This is a
        destructive action clearing all state and removing the flag from the
        set of "known features".
        """
        return self._backend.remove(feature)

    def _to_flipper_id(self, object) -> str:
        if hasattr(object, 'get_flipper_id'):
            return object.get_flipper_id()

        type_name = object.__class__.__name__
        object_id = None

        if hasattr(object, 'flipper_id'):
            object_id = object.flipper_id
        elif hasattr(object, 'pk'):
            object_id = object.pk
        elif hasattr(object, 'id'):
            object_id = object.id
        else:
            # TODO: check if this is stable over time
            object_id = hash(object)
        
        return f"{type_name};{object_id}"



import json

from flippy.backends import BaseBackend
from flippy.core import Feature, FeatureEncoder, FeatureName, Gate
from flippy.exceptions import FeatureNotFound

from django.core.exceptions import ImproperlyConfigured
try:
    from flippy.models import FlippyFeature, FlippyActorGate, FlippyGroupGate
except ImproperlyConfigured:
    pass

from django.core.exceptions import ValidationError
from django.db.utils import IntegrityError


class DjangoBackend(BaseBackend):
    """
    A backend implemented as Django models.
    """
    def features(self) -> set[FeatureName]:
        "Get the set of known features."
        return set([f.key for f in FlippyFeature.objects.all()])

    def add(self, feature: FeatureName) -> bool:
        "Add a feature to the set of known features."
        try:
            FlippyFeature(key=feature).save()
            return True
        except IntegrityError:
            return False

    def remove(self, feature: FeatureName) -> bool:
        "Remove a feature from the set of known features."
        try:
            feature = FlippyFeature.objects.get(key=feature)
            feature.delete()
            return True
        except FlippyFeature.DoesNotExist:
            return False

    def clear(self, feature: FeatureName) -> bool:
        "Clear all gate values for a feature."
        try:
            feature = FlippyFeature.objects.get(key=feature)
            feature.clear()
            return True
        except FlippyFeature.DoesNotExist:
            return False

    def get(self, feature: FeatureName) -> Feature:
        "Get all gate values for a feature."
        try:
            feature = FlippyFeature.objects.get(key=feature)
            return feature.as_feature()
        except FlippyFeature.DoesNotExist:
            raise FeatureNotFound(feature)

    def enable(self, feature: FeatureName, gate: Gate, thing: str | int | None=None) -> bool:
        "Enable a gate for a thing."
        try:
            feature = FlippyFeature.objects.get(key=feature)
        except FlippyFeature.DoesNotExist:
            return False

        match gate:
            case Gate.Boolean:
                feature.boolean = True
                feature.save()
                return True
            case Gate.Actors:
                try:
                    feature.enabled_actors.create(key=thing)
                    return True
                except IntegrityError:
                    return False
            case Gate.Groups:
                try:
                    feature.enabled_groups.create(key=thing)
                    return True
                except IntegrityError:
                    return False
            case Gate.PercentageOfActors:
                try:
                    feature.percentage_of_actors = thing
                    feature.save()
                    return True
                except ValidationError:
                    return False
            case Gate.PercentageOfTime:
                try:
                    feature.percentage_of_time = thing
                    feature.save()
                    return True
                except ValidationError:
                    return False
            case Gate.Expression:
                raise NotImplementedError("ExpressionGate isn't supported")
            case _:
                raise ValueError(f"{gate} is not a known gate type")

    def disable(self, feature: FeatureName, gate: Gate, thing: str | int | None=None) -> bool:
        "Disable a gate for a thing."
        try:
            feature = FlippyFeature.objects.get(key=feature)
        except FlippyFeature.DoesNotExist:
            return False

        match gate:
            case Gate.Boolean:
                feature.boolean = False
                feature.save()
                return True
            case Gate.Actors:
                try:
                    feature.enabled_actors.get(key=thing).delete()
                    return True
                except FlippyActorGate.DoesNotExist:
                    return False
            case Gate.Groups:
                try:
                    feature.enabled_groups.get(key=thing).delete()
                    return True
                except FlippyGroupGate.DoesNotExist:
                    return False
            case Gate.PercentageOfActors:
                try:
                    feature.percentage_of_actors = None
                    feature.save()
                    return True
                except ValidationError:
                    return False
            case Gate.PercentageOfTime:
                try:
                    feature.percentage_of_time = None
                    feature.save()
                    return True
                except ValidationError:
                    return False
            case Gate.Expression:
                raise NotImplementedError("ExpressionGate isn't supported")
            case _:
                raise ValueError(f"{gate} is not a known gate type")

    def get_multi(self, features: list[FeatureName]) -> list[Feature]:
        "Get all gate values for several features at once."
        features = FlippyFeature.objects.filter(key__in=features)
        return [f.as_feature() for f in features]

    def get_all(self) -> list[Feature]:
        "Get all gate values for all features at once."
        features = FlippyFeature.objects.all()
        return [f.as_feature() for f in features]

    def to_json(self) -> str:
        "Produce a JSON-formatted string containing state for all features."
        return super().to_json()

    def from_json(self, new_state: str) -> None:
        "Clear current state and replace with state from a JSON-formatted string."
        return self._from_json_naive(new_state)

    def _from_json_naive(self, new_state: str) -> None:
        FlippyFeature.objects.all().delete()
        features_raw = json.loads(new_state)
        for value in features_raw.values():
            f = Feature.from_api(value)
            # FlippyFeature.from_feature saves the object
            ff = FlippyFeature.from_feature(f)



from dataclasses import dataclass, field
from enum import Enum
import json
from typing import Literal, NewType

from flippy.gates import (ActorsGate, BooleanGate, ExpressionGate, GroupsGate,
                          PercentageOfActorsGate, PercentageOfTimeGate)

FeatureName = NewType('FeatureName', str)
"The (string) name of a feature."

FlagState = Literal['on', 'conditional', 'off']


class Gate(Enum):
    Boolean = 'boolean'
    Actors = 'actors'
    Groups = 'groups'
    PercentageOfActors = 'percentage_of_actors'
    PercentageOfTime = 'percentage_of_time'
    Expression = 'expression'


@dataclass
class Feature:
    key: FeatureName

    @property
    def state(self) -> FlagState:
        if self.boolean_gate.value:
            return 'on'
        if (self.actors_gate.value == []
            and self.groups_gate.value == []
            and self.percentage_of_actors_gate.value == None
            and self.percentage_of_time_gate.value == None
            and self.expression_gate.value == None):
            return 'off'
        return 'conditional'

    boolean_gate: BooleanGate = field(default_factory=BooleanGate)
    actors_gate: ActorsGate = field(default_factory=ActorsGate)
    groups_gate: GroupsGate = field(default_factory=GroupsGate)
    percentage_of_actors_gate: PercentageOfActorsGate = field(default_factory=PercentageOfActorsGate)
    percentage_of_time_gate: PercentageOfTimeGate = field(default_factory=PercentageOfTimeGate)
    expression_gate: ExpressionGate = field(default_factory=ExpressionGate)

    def to_api(self):
        return {
            'key': self.key,
            'state': self.state,
            'gates': [
                self.boolean_gate.to_api(),
                self.actors_gate.to_api(),
                self.groups_gate.to_api(),
                self.percentage_of_actors_gate.to_api(),
                self.percentage_of_time_gate.to_api(),
                self.expression_gate.to_api(),
            ]
        }
    
    @classmethod
    def from_api(cls, api_payload: dict):
        self = cls(api_payload['key'])
        for gate in api_payload['gates']:
            gateValue = gate['value']
            match gate['key']:
                case 'boolean':
                    self.boolean_gate.value = gateValue
                case 'actors':
                    self.actors_gate.value = gateValue
                case 'groups':
                    self.groups_gate.value = gateValue
                case 'percentage_of_actors':
                    self.percentage_of_actors_gate.value = None if gateValue is None else int(gateValue)
                case 'percentage_of_time':
                    self.percentage_of_time_gate.value = None if gateValue is None else int(gateValue)
                case 'expression':
                    self.expression_gate.value = gateValue
                case _:
                    raise ValueError(f"{gate} is not a known gate type")
        return self

    def __eq__(self, other):
        if not isinstance(other, Feature):
            return False
        
        return all([
            self.key == other.key,
            self.boolean_gate == other.boolean_gate,
            self.actors_gate == other.actors_gate,
            self.groups_gate == other.groups_gate,
            self.percentage_of_actors_gate == other.percentage_of_actors_gate,
            self.percentage_of_time_gate == other.percentage_of_time_gate,
            self.expression_gate == other.expression_gate,
        ])


class FeatureEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Feature):
            return obj.to_api()
        return json.JSONEncoder.default(self, obj)



import typing

if typing.TYPE_CHECKING:
    from flippy.core import FeatureName


class FeatureNotFound(Exception):
    def __init__(self, feature: 'FeatureName', parent_exception: Exception | None = None, *args: object) -> None:
        self.parent_exception = parent_exception
        super().__init__(feature, *args)


class GroupNotRegistered(Exception): pass
class PercentageInvalid(Exception): pass
class FlipperIdInvalid(Exception): pass
class NameInvalid(Exception): pass



from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models

from flippy.core import Feature


class FlippyFeature(models.Model):
    key = models.CharField(max_length=150, unique=True)
    boolean = models.BooleanField(null=True)
    percentage_of_actors = models.IntegerField(
        null=True,
        default=None,
        validators=[
            MinValueValidator(0),
            MaxValueValidator(100),
        ]
    )
    percentage_of_time = models.IntegerField(
        null=True,
        default=None,
        validators=[
            MinValueValidator(0),
            MaxValueValidator(100),
        ]
    )

    def __str__(self):
        return self.key
    
    def clear(self):
        self.boolean = None
        self.percentage_of_actors = None
        self.percentage_of_time = None
        self.save()
        self.enabled_actors.all().delete()
        self.enabled_groups.all().delete()
    
    def as_feature(self):
        f = Feature(self.key)
        f.boolean_gate.value = self.boolean
        f.actors_gate.value = [a.key for a in self.enabled_actors.all()]
        f.groups_gate.value = [g.key for g in self.enabled_groups.all()]
        f.percentage_of_actors_gate.value = self.percentage_of_actors
        f.percentage_of_time_gate.value = self.percentage_of_time
        return f
    
    @classmethod
    def from_feature(cls, feature: Feature):
        f = cls(key=feature.key)
        f.save()    # to get an ID assigned
        f.boolean = feature.boolean_gate.value
        for a in feature.actors_gate.value:
            f.enabled_actors.create(key=a)
        for g in feature.groups_gate.value:
            f.enabled_groups.create(key=g)
        f.percentage_of_actors = feature.percentage_of_actors_gate.value
        f.percentage_of_time = feature.percentage_of_time_gate.value
        f.save()
        return f


class FlippyActorGate(models.Model):
    key = models.CharField(max_length=150)
    feature = models.ForeignKey(
        FlippyFeature,
        on_delete=models.CASCADE,
        related_name='enabled_actors',
    )

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['key', 'feature'], name='actor_feature_uniq')
        ]
    
    def __str__(self):
        return self.key


class FlippyGroupGate(models.Model):
    key = models.CharField(max_length=150)
    feature = models.ForeignKey(
        FlippyFeature,
        on_delete=models.CASCADE,
        related_name='enabled_groups',
    )

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['key', 'feature'], name='group_feature_uniq')
        ]
    
    def __str__(self):
        return self.key



from flippy.backends.base import BaseBackend
from flippy.backends.django import DjangoBackend
from flippy.backends.flipper_cloud import FlipperCloudBackend
from flippy.backends.memory import MemoryBackend

__all__ = [
    BaseBackend,
    DjangoBackend,
    FlipperCloudBackend,
    MemoryBackend,
]



import logging

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from flippy.backends import DjangoBackend

logger = logging.getLogger(__name__)

# get class
try:
    _backend = settings.FLIPPY_BACKEND
    logger.debug('Flippy backend found in settings: "%s"', _backend)

    # get a class from the string
    try:
        if '.' not in _backend:
            logger.debug('Assuming this is a built-in backend')
            _backend = f'flippy.backends.{_backend}'
        _class = import_string(_backend)

    except ImportError:
        raise ImproperlyConfigured(
            f"Backend `{_backend}` was not found; try passing a fully-qualified path as a dotted string."
        )

except AttributeError:
    # default if nothing is configured
    logger.debug('Did not find FLIPPY_BACKEND in settings; defaulting to DjangoBackend')
    _class = DjangoBackend

# check for settings
try:
    _args = settings.FLIPPY_ARGS
except AttributeError:
    logger.debug('No FLIPPY_ARGS found in settings; defaulting to empty list')
    _args = []

# build it
if isinstance(_args, (list, tuple)):
    flippy_backend = _class(*_args)
elif isinstance(_args, dict):
    flippy_backend = _class(**_args)
elif isinstance(_args, str):
    flippy_backend = _class(_args)
else:
    logger.warn("settings.FLIPPY_ARGS had content, but it wasn't a list, tuple, dict, or str")
    flippy_backend = _class()



from django.core.management.base import BaseCommand, CommandError
from flippy.config import flippy_backend
from flippy.backends import FlipperCloudBackend
from os import environ

try:
    TOKEN = environ['FLIPPER_CLOUD_TOKEN']
except KeyError:
    TOKEN = None


class Command(BaseCommand):
    help = "Sync from Flipper Cloud to another backend"

    def handle(self, *args, **options):
        if not TOKEN:
            raise CommandError(
                "FLIPPER_CLOUD_TOKEN must be set in the environment for sync to operate"
            )

        if isinstance(flippy_backend, FlipperCloudBackend):
            raise CommandError(
                "Will not sync from Flipper Cloud back to Flipper Cloud"
            )

        source = FlipperCloudBackend(TOKEN)
        flippy_backend.from_json(source.to_json())

        self.stdout.write(
            self.style.SUCCESS(
                f"Completed sync from {source.__class__.__name__} "
                f"to {flippy_backend.__class__.__name__}"
            )
        )



import platform
import sys
from json.decoder import JSONDecodeError

import httpx

from flippy.backends import BaseBackend
from flippy.core import Feature, FeatureName, Gate
from flippy.exceptions import (FeatureNotFound, FlipperIdInvalid,
                               GroupNotRegistered, NameInvalid,
                               PercentageInvalid)

FLIPPER_CLOUD_BASE_URL = 'https://www.flippercloud.io/adapter'
SYSTEM_PLATFORM = f"{platform.machine() or 'unknown'}-{platform.system() or 'unknown'}"
HEADERS = {
    'Content-Type': 'application/json',
    'Accept': 'application/json',
    'User-Agent': "django-flippy v0.0.1",
    "Client-Language": "python",
    "Client-Language-Version": sys.version,
    "Client-Platform": SYSTEM_PLATFORM,
    # TODO: others from https://github.com/flippercloud/flipper/blob/main/lib/flipper/adapters/http/client.rb
    # "Client-Engine": "",
    # "Client-Pid": Process.pid.to_s,
    # "Client-Thread": Thread.current.object_id.to_s,
    # "Client-Hostname": Socket.gethostname,
}


# As of 2023-08-26, there are many cases where Flipper Cloud responds
# with success even if the feature doesn't exist or nothing was changed.
# This makes many requests idempotent, but also makes it hard to surface
# problems to the consumer. 🤷

class FlipperCloudBackend(BaseBackend):
    """
    Get flags direct from Flipper Cloud.
    
    Note that this is not the preferred implementation, as it'll cause a service
    to service call on every invocation. However, it's good for scaffolding and
    kicking tires.
    """
    def __init__(self, token: str):
        if not token:
            raise ValueError('must pass a Flipper Cloud token')

        self.client = httpx.Client(
            base_url=FLIPPER_CLOUD_BASE_URL,
            headers=HEADERS | { "Flipper-Cloud-Token": token },
        )

    def features(self) -> set[FeatureName]:
        "Get the set of known features."
        r = self.client.get('/features?exclude_gate_names=true')
        self._raise_from_cloud(r)

        return {f['key'] for f in r.json()['features']}

    def add(self, feature: FeatureName) -> bool:
        "Add a feature to the set of known features."
        body = { 'name': feature }
        r = self.client.post('/features', json=body)
        self._raise_from_cloud(r, { 'feature': feature })
        return r.is_success

    def remove(self, feature: FeatureName) -> bool:
        "Remove a feature from the set of known features."
        r = self.client.delete(f'/features/{feature}')
        self._raise_from_cloud(r, { 'feature': feature })
        return r.is_success

    def clear(self, feature: FeatureName) -> bool:
        "Clear all gate values for a feature."
        r = self.client.delete(f'features/{feature}/clear')
        self._raise_from_cloud(r, { 'feature': feature })
        return r.is_success

    def get(self, feature: FeatureName) -> Feature:
        "Get all gate values for a feature."
        r = self.client.get(f'/features/{feature}')

        self._raise_from_cloud(r, { 'feature': feature })
        return Feature.from_api(r.json())        

    def enable(self, feature: FeatureName, gate: Gate, thing: str | int | None=None) -> bool:
        "Enable a gate for a thing."
        body = self._body_for_gate(gate, thing)
        qs = { 'allow_unregistered_groups': 'true' } if gate == Gate.Groups else {}
        r = self.client.post(
            f'/features/{feature}/{gate.value}',
            json=body,
            params=qs,
        )
        
        self._raise_from_cloud(r, { 'feature': feature })
        return r.is_success

    def disable(self, feature: FeatureName, gate: Gate, thing: str | int | None=None) -> bool:
        "Disable a gate for a thing."
        body = self._body_for_gate(gate, thing)
        qs = { 'allow_unregistered_groups': 'true' } if gate == Gate.Groups else {}
        # httpx doesn't accept `json` on .delete(), but the body is a required part
        # of some disable requests
        req = self.client.build_request(
            'DELETE',
            f'/features/{feature}/{gate.value}',
            json=body,
            params=qs,
        )
        r = self.client.send(req)
        self._raise_from_cloud(r, { 'feature': feature })

        return r.is_success

    def get_multi(self, features: list[FeatureName]) -> list[Feature]:
        "Get all gate values for several features at once."
        qs = {
            'exclude_gate_names': 'true',
            'keys': ','.join(features),
        }
        r = self.client.get('/features', params=qs)
        self._raise_from_cloud(r)

        return [Feature.from_api(f) for f in r.json()['features']]

    def get_all(self) -> list[Feature]:
        "Get all gate values for all features at once."
        qs = {
            'exclude_gate_names': 'true',
        }
        r = self.client.get('/features', params=qs)
        self._raise_from_cloud(r)

        return [Feature.from_api(f) for f in r.json()['features']]

    def _body_for_gate(self, gate: Gate, thing: str | int | None) -> dict:
        match gate:
            case Gate.Boolean:
                return {}
            case Gate.Actors:
                return { 'flipper_id': str(thing) }
            case Gate.Groups:
                return { 'name': str(thing) }
            case Gate.PercentageOfActors | Gate.PercentageOfTime:
                return { 'percentage': str(thing) }
            case Gate.Expression:
                raise NotImplementedError("ExpressionGate isn't supported")
            case _:
                raise ValueError(f"{gate} is not a known gate type")

    def _raise_from_cloud(self, response: httpx.Response, extra: dict = {}):
        # https://www.flippercloud.io/docs/api#error-code-reference
        if response.is_client_error or response.is_success:
            try:
                result = response.json()
            except JSONDecodeError:
                result = {}
            if 'code' not in result:
                return
            
            match result['code']:
                case 1:
                    raise FeatureNotFound(extra.get('feature', ''))
                case 2:
                    raise GroupNotRegistered(result.get('message', ''))
                case 3:
                    raise PercentageInvalid(result.get('message', ''))
                case 4:
                    raise FlipperIdInvalid(result.get('message', ''))
                case 5:
                    raise NameInvalid(result.get('message', ''))

        response.raise_for_status()

    def to_json(self) -> str:
        "Produce a JSON-formatted string containing state for all features."
        return super().to_json()

    def from_json(self, new_state: str) -> None:
        "Clear current state and replace with state from a JSON-formatted string."
        raise NotImplementedError()



import random
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, NewType
from zlib import crc32

if TYPE_CHECKING:
    from flippy.core import FeatureName

Percentage = NewType('Percentage', int)


@dataclass
class BooleanGate:
    value: bool | None = None

    def to_api(self):
        return {
            'key': 'boolean',
            'name': 'boolean',
            'value': self.value,
        }

    def is_open(self, target: str, feature: 'FeatureName'):
        return bool(self.value)

    def __eq__(self, other):
        if not isinstance(other, BooleanGate):
            return False
        
        return self.value == other.value

@dataclass
class ActorsGate:
    value: list[str]

    def __init__(self, value: list[str]=None):
        self.value = value or []

    def to_api(self):
        return {
            'key': 'actors',
            'name': 'actor',
            'value': self.value,
        }

    def is_open(self, target: str, feature: 'FeatureName'):
        return target in self.value

    def __eq__(self, other):
        if not isinstance(other, ActorsGate):
            return False

        if self.value is None or other.value is None:
            return False
        
        return sorted(self.value) == sorted(other.value)

@dataclass
class GroupsGate:
    value: list[str]

    def __init__(self, value: list[str]=None):
        self.value = value or []

    def to_api(self):
        return {
            'key': 'groups',
            'name': 'group',
            'value': self.value,
        }

    def is_open(self, target: str, feature: 'FeatureName'):
        return target in self.value

    def __eq__(self, other):
        if not isinstance(other, GroupsGate):
            return False
        
        if self.value is None or other.value is None:
            return False
        
        return sorted(self.value) == sorted(other.value)

@dataclass
class PercentageOfActorsGate:
    value: Percentage | None = None

    def to_api(self):
        return {
            'key': 'percentage_of_actors',
            'name': 'percentage_of_actors',
            'value': str(self.value) if self.value is not None else None,
        }

    def is_open(self, target: str, feature: 'FeatureName'):
        if self.value is None:
            return False
        if self.value == 100:
            return True

        # Roughly we need to:
        # - stably mash the target + feature name into a number 0-100
        # - compare that result to self.value
        # - return (our number < self.value)
        # The reason for strict less-than is so that if self.value is set to
        # 0, no actors slip through.
        feature_hash = crc32(feature.encode())
        target_hash = crc32(target.encode(), feature_hash)
        target_value = round(100 * target_hash / 2**32)
        return target_value < self.value

    def __eq__(self, other):
        if not isinstance(other, PercentageOfActorsGate):
            return False
        
        return self.value == other.value

@dataclass
class PercentageOfTimeGate:
    value: Percentage | None = None

    def to_api(self):
        return {
            'key': 'percentage_of_time',
            'name': 'percentage_of_time',
            'value': str(self.value) if self.value is not None else None,
        }

    def is_open(self, target: str, feature: 'FeatureName'):
        if self.value is None:
            return False

        r = random.randint(0, 100)
        return self.value >= r

    def __eq__(self, other):
        if not isinstance(other, PercentageOfTimeGate):
            return False
        
        return self.value == other.value

@dataclass
class ExpressionGate:
    value: Any = None # TODO

    def to_api(self):
        return {
            'key': 'expression',
            'name': 'expression',
            'value': self.value,
        }

    def is_open(self, target: str, feature: 'FeatureName'):
        return False

    def __eq__(self, other):
        if not isinstance(other, ExpressionGate):
            return False

        return self.value == other.value



import copy
import json

from flippy.backends import BaseBackend
from flippy.core import Feature, FeatureEncoder, FeatureName, Gate
from flippy.exceptions import FeatureNotFound


class MemoryBackend(BaseBackend):
    """A memory-only implementation of Flippy."""
    def __init__(self):
        self._features: dict[FeatureName: Feature] = {}

    def features(self) -> set[FeatureName]:
        "Get the set of known features."
        return self._features.keys()

    def add(self, feature: FeatureName) -> bool:
        "Add a feature to the set of known features."
        if feature in self._features:
            return False
        self._features[feature] = Feature(key=feature)
        return True

    def add_feature(self, feature: Feature) -> bool:
        "Add a fully hydrated feature, such as from an API call."
        if feature.key in self._features:
            return False
        self._features[feature.key] = feature
        return True

    def remove(self, feature: FeatureName) -> bool:
        "Remove a feature from the set of known features."
        try:
            self._features.pop(feature)
            return True
        except KeyError:
            return False

    def clear(self, feature: FeatureName) -> bool:
        "Clear all gate values for a feature."
        if feature not in self._features:
            return False
        self._features[feature] = Feature(key=feature)

    def get(self, feature: FeatureName) -> Feature:
        "Get all gate values for a feature."
        try:
            return copy.deepcopy(self._features[feature])
        except KeyError:
            raise FeatureNotFound(feature)

    def enable(self, feature: FeatureName, gate: Gate, thing: str | int | None=None) -> bool:
        "Enable a gate for a thing."
        feature = self._features[feature]

        match gate:
            case Gate.Boolean:
                feature.boolean_gate.value = True
                return True
            case Gate.Actors:
                if thing and thing not in feature.actors_gate.value:
                    feature.actors_gate.value.append(thing)
                    return True
                return False
            case Gate.Groups:
                if thing and thing not in feature.groups_gate.value:
                    feature.groups_gate.value.append(thing)
                    return True
                return False
            case Gate.PercentageOfActors:
                # TODO: check if thing is int
                feature.percentage_of_actors_gate.value = thing
                return True
            case Gate.PercentageOfTime:
                # TODO: check if thing is int
                feature.percentage_of_time_gate.value = thing
                return True
            case Gate.Expression:
                raise NotImplementedError("ExpressionGate isn't supported")
            case _:
                raise ValueError(f"{gate} is not a known gate type")

    def disable(self, feature: FeatureName, gate: Gate, thing: str | int | None=None) -> bool:
        "Disable a gate for a thing."
        feature = self._features[feature]

        match gate:
            case Gate.Boolean:
                feature.boolean_gate.value = False
                return True
            case Gate.Actors:
                if thing and thing in feature.actors_gate.value:
                    feature.actors_gate.value.remove(thing)
                    return True
                return False
            case Gate.Groups:
                if thing and thing in feature.groups_gate.value:
                    feature.groups_gate.value.remove(thing)
                    return True
                return False
            case Gate.PercentageOfActors:
                feature.percentageofactors_gate.value = None
                return True
            case Gate.PercentageOfTime:
                feature.percentageoftime_gate.value = None
                return True
            case Gate.Expression:
                raise NotImplementedError("ExpressionGate isn't supported")
            case _:
                raise ValueError(f"{gate} is not a known gate type")

    def get_multi(self, features: list[FeatureName]) -> list[Feature]:
        "Get all gate values for several features at once."
        return super().get_multi(features)

    def get_all(self) -> list[Feature]:
        "Get all gate values for all features at once."
        return super().get_all()

    def to_json(self) -> str:
        "Produce a JSON-formatted string containing state for all features."
        return json.dumps(self._features, cls=FeatureEncoder, separators=(',', ':'))

    def from_json(self, new_state: str) -> None:
        "Clear current state and replace with state from a JSON-formatted string."
        self._features = {}
        features_raw = json.loads(new_state)
        for k, v in features_raw.items():
            self._features[k] = Feature.from_api(v)



from django.core.exceptions import ImproperlyConfigured
from django.utils.functional import lazy

from flippy import Flippy
from flippy.config import flippy_backend


class FeatureContext:
    def __init__(self, flippy: Flippy, name, request):
        self.flippy = flippy
        self.name = name
        self.request = request
    
    def for_user(self) -> bool:
        try:
            user = self.request.user
        except AttributeError:
            raise ImproperlyConfigured(
                "request.user doesn't exist; make sure "
                "django.contrib.auth.context_processors.auth is on in your settings."
            )
        return self.flippy.is_enabled(self.name, user)
    
    def globally(self) -> bool:
        return self.flippy.is_enabled(self.name)
    
    def __bool__(self):
        raise NotImplementedError(
            f"Asking for `flippy.{self.name}` isn't implemented. "
            "(It seems like a trap / easy typo to make, with unintended consequences.) "
            f"You probably want `flippy.{self.name}.globally` instead."
        )


class FlippyContext:
    def __init__(self, request):
        self.flippy = Flippy(flippy_backend)
        self.request = request
    
    def __getattr__(self, name: str) -> FeatureContext:
        if self.flippy.feature_exists(name):
            return FeatureContext(self.flippy, name, self.request)
        raise AttributeError(name=name, obj=self)


def processor(request):
    # Only instantiate the FlippyContext if someone actually requests it.
    # Because you have to call `flippy.<feature_name>`, the FeatureContext
    # is essentially already lazy.
    c = lambda: FlippyContext(request)
    return {
        'flippy': lazy(c, FlippyContext),
    }



# Generated by Django 4.2.4 on 2023-08-30 13:11

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='FlippyFeature',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=150, unique=True)),
                ('boolean', models.BooleanField(null=True)),
                ('percentage_of_actors', models.IntegerField(default=None, null=True, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(100)])),
                ('percentage_of_time', models.IntegerField(default=None, null=True, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(100)])),
            ],
        ),
        migrations.CreateModel(
            name='FlippyGroupGate',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=150)),
                ('feature', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='enabled_groups', to='flippy.flippyfeature')),
            ],
        ),
        migrations.CreateModel(
            name='FlippyActorGate',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=150)),
                ('feature', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='enabled_actors', to='flippy.flippyfeature')),
            ],
        ),
        migrations.AddConstraint(
            model_name='flippygroupgate',
            constraint=models.UniqueConstraint(fields=('key', 'feature'), name='group_feature_uniq'),
        ),
        migrations.AddConstraint(
            model_name='flippyactorgate',
            constraint=models.UniqueConstraint(fields=('key', 'feature'), name='actor_feature_uniq'),
        ),
    ]



import importlib.metadata
from flippy.flippy import Flippy

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except:
    __version__ = "0.0.1-dev"


__all__ = [
    Flippy,
]



from flippy import Flippy
from flippy.config import flippy_backend


def flippy_middleware(get_response):
    flippy = Flippy(flippy_backend)

    def middleware(request):
        request.flippy = flippy
        response = get_response(request)
        return response
    return middleware
